// SPDX-License-Identifier: Apache-2.0
pragma solidity ^0.8.0;

/// @author thirdweb

/**
 *  Group together arbitrary ERC20, ERC721 and ERC1155 tokens into a single bundle.
 *
 *  The `Token` struct is a generic type that can describe any ERC20, ERC721 or ERC1155 token.
 *  The `Bundle` struct is a data structure to track a group/bundle of multiple assets i.e. ERC20,
 *  ERC721 and ERC1155 tokens, each described as a `Token`.
 *
 *  Expressing tokens as the `Token` type, and grouping them as a `Bundle` allows for writing generic
 *  logic to handle any ERC20, ERC721 or ERC1155 tokens.
 */

interface ITokenBundle {
    /// @notice The type of assets that can be wrapped.
    enum TokenType {
        ERC20,
        ERC721,
        ERC1155
    }

    /**
     *  @notice A generic interface to describe any ERC20, ERC721 or ERC1155 token.
     *
     *  @param assetContract The contract address of the asset.
     *  @param tokenType     The token type (ERC20 / ERC721 / ERC1155) of the asset.
     *  @param tokenId       The token Id of the asset, if the asset is an ERC721 / ERC1155 NFT.
     *  @param totalAmount   The amount of the asset, if the asset is an ERC20 / ERC1155 fungible token.
     */
    struct Token {
        address assetContract;
        TokenType tokenType;
        uint256 tokenId;
        uint256 totalAmount;
    }

    /**
     *  @notice An internal data structure to track a group / bundle of multiple assets i.e. `Token`s.
     *
     *  @param count    The total number of assets i.e. `Token` in a bundle.
     *  @param uri      The (metadata) URI assigned to the bundle created
     *  @param tokens   Mapping from a UID -> to a unique asset i.e. `Token` in the bundle.
     */
    struct BundleInfo {
        uint256 count;
        string uri;
        mapping(uint256 => Token) tokens;
    }
}
